import numpy as np
import pandas as pd
import math
import os

from ..constants import ANNOTATED_SUFFIX, ANNOTATION_COLUMN, ANNOTATOR_COLUMN


class AnnotationService:
    def __init__(self):
        self.df = None
        self.id_column = None
        
        self.current_csv_path = ''
        self.csv_annotated = self.get_annotated_csv_path(self.current_csv_path)
        

    def get_annotated_csv_path(self, csv_path):
        return csv_path.replace('.csv', ANNOTATED_SUFFIX + '.csv')
        
    def create_annotated_csv(self):
        self.df = pd.read_csv(self.current_csv_path)
        self.df[ANNOTATOR_COLUMN] = np.nan
        self.df[ANNOTATION_COLUMN] = np.nan
        self.df.to_csv(self.csv_annotated, index=False)
    
    def load_annotated_csv(self):
        self.df = pd.read_csv(self.csv_annotated)


    def get_images_from_rows(self, selected_rows):
        images = []
        for ind, row in selected_rows.iterrows():
            images.append({
                'id': row[self.id_column],
                'index': ind,
                'value': row[ANNOTATION_COLUMN] if not math.isnan(row[ANNOTATION_COLUMN]) else 'None'
            })
        return images


    def get_first_index_not_fully_annotated_page(self, offset):
        df_not_annotated = self.df[self.df[ANNOTATION_COLUMN].isna()]
        
        if len(df_not_annotated) == 0:
            return len(self.df)
        
        first_not_annotated = df_not_annotated.iloc[0]
        first_index = self.df[self.df[self.id_column] == first_not_annotated[self.id_column]].index[0]
        start = int(offset * np.floor(first_index / offset))
        
        return start
    
    def get_n_annotations(self, first, offset):
        if first == -1:
            first = self.get_first_index_not_fully_annotated_page(offset)
        selected_rows = self.df[first: first + offset]
        images = self.get_images_from_rows(selected_rows)
        
        return images

    
    def get_annotations(self, csv_path, first, offset, id_column):
        if csv_path != self.current_csv_path:
            self.current_csv_path = csv_path
            self.csv_annotated = self.get_annotated_csv_path(csv_path)
            
            if os.path.exists(self.csv_annotated):
                self.load_annotated_csv()
            elif os.path.exists(self.current_csv_path):
                self.create_annotated_csv()
            else:
                return { 'error': f'CSVs not found: {csv_path} and {self.csv_annotated}' }, 200
            
        if id_column is None or id_column not in self.df:
            return { 'error': f'Column {id_column} not in the csv.' }, 200
        if id_column != self.id_column:
            self.id_column = id_column
            
        images = self.get_n_annotations(first, offset)
        
        return {
            'images': images,
            'processed': len(self.df[~self.df[ANNOTATION_COLUMN].isna()]),
            'total': len(self.df)
        }, 200


    def annotate(self, annotations, annotator):
        for idx, is_valid in annotations.items():
            self.df.loc[self.df[self.id_column] == idx, ANNOTATOR_COLUMN] = annotator
            self.df.loc[self.df[self.id_column] == idx, ANNOTATION_COLUMN] = is_valid
        self.df.to_csv(self.csv_annotated, index=False)



from flask import Flask, request, send_file
from flask_restx import Resource
import io
from PIL import Image

from . import api
from .utils import sha_generator, print_important
from ..constants import PASSWORD_ERROR

from ..services import AnnotationService, ImageService


SHA_HASH = sha_generator()
print_important("Password to add in the frontend options:", important_msg=SHA_HASH)


annotation_service = AnnotationService()
image_service = ImageService()


def send_numpy_image(image):
    img = Image.fromarray(image.astype('uint8'))
    file_object = io.BytesIO()
    img.save(file_object, 'PNG')
    file_object.seek(0)
            
    return send_file(file_object, mimetype='image/PNG')



@api.route('/get-annotations', methods=['POST'])
class GetAnnotations(Resource):
    def post(self):        
        body = request.json

        first = body['first'] if 'first' in body else -1
        offset = int(body['offset']) if 'offset' in body else 8
        sha_hash = body['sha'] if 'sha' in body else ''
        csv_path = body['csv_path'] if 'csv_path' in body else None
        images_folder = body['images_folder'] if 'images_folder' in body else ''
        masks_folder = body['masks_folder'] if 'masks_folder' in body else None
        with_masks = body['with_masks'] if 'with_masks' in body else False
        id_column = body['id_column_name'] if 'id_column_name' in body else None
        image_ext = body['image_ext'] if 'image_ext' in body else ''
        mask_ext = body['mask_ext'] if 'mask_ext' in body else ''
        
        if sha_hash != SHA_HASH:
            return { 'error': PASSWORD_ERROR }, 200
        
        check_img = image_service.set_constants(images_folder, image_ext, masks_folder, mask_ext, with_masks)
        
        if check_img is not None:
            return check_img
        return annotation_service.get_annotations(csv_path, first, offset, id_column)
        
        
@api.route('/get-image', methods=['GET'])
class GetImage(Resource):
    def get(self):
        image_id = request.args.get('id', '')
        sha_hash = request.args.get('sha', '')
        
        if sha_hash != SHA_HASH:
            return { 'error': PASSWORD_ERROR }, 200

        image = image_service.build_image(image_id)
        
        if type(image) is tuple:
            return image
        return send_numpy_image(image)
        
        
@api.route('/annotate', methods=['POST'])
class Annotate(Resource):
    def post(self):
        body = request.json
        annotator, annotations, sha_hash = body['annotator'], body['annotations'], body['sha']
        
        if sha_hash != SHA_HASH:
            return { 'error': PASSWORD_ERROR }, 200
        
        annotation_service.annotate(annotations, annotator)
        
        return { 'msg': 'Annotation done!' }, 200



from flask import Flask
from flask_cors import CORS
from werkzeug.utils import cached_property

from .api import annotation_api


FRONTEND_FOLDER = '../front/build/'
app = Flask(__name__, static_url_path='', static_folder=FRONTEND_FOLDER)

API_PREFIX = '/api'
app.register_blueprint(annotation_api, url_prefix=API_PREFIX)
CORS(app)

import sys, os,  pathlib
@app.route('/')
def root():
    return app.send_static_file('index.html')



import cv2
import numpy as np
from PIL import Image


MAX_SIZE_IMG = 600


"""
    ========== Functions to customize if necessary ==========
    Images are built and sent to frontend thanks to build_image function.
    
    Inputs: local paths to images and optionally masks.
    Output: numpy image
"""
def build_image(img_path, mask_path=None):
    img = open_image(img_path)
    
    if mask_path is None:
        return resize_image(img)
    
    mask = open_image(mask_path)
    composition = compose_img(img, mask)
    img = np.concatenate((img, composition), axis=1)
    
    return resize_image(img)


def compose_img(img, mask):    
    bg = np.zeros_like(img)
    bg[:,:,1] = 1.0

    mask = cv2.resize(mask, (img.shape[1], img.shape[0]), interpolation=cv2.INTER_CUBIC)
    mask, img = mask / 255, img / 255
    fg = mask * img
    bg = (1-mask) * bg
    cutout = np.array(255 * (fg + bg), dtype=np.uint8)
    
    return cutout


def open_image(path):
    return np.array(Image.open(path).convert('RGBA'))


def resize_image(img):
    h, w, _ = img.shape
    ratio = min(MAX_SIZE_IMG/h, MAX_SIZE_IMG/w)
    img = cv2.resize(img, None, fx=ratio, fy=ratio, interpolation=cv2.INTER_AREA)
    return img



import random as rd
import string
from hashlib import sha256

# rd.seed(0)    # For dev only
def sha_generator():
    word = "".join([rd.choice(string.ascii_letters) for i in range(10)])
    return sha256(word.encode('ascii')).hexdigest()


BOLD = '\033[1m'
END = '\033[0m'
UNDERLINE = '\033[4m'
GREEN = '\033[92m'
def print_important(message, important_msg=''):
    dash_line = '=' * 50
    message = UNDERLINE + GREEN + message + END
    important_msg = f'\n\n==>\t{BOLD}{important_msg}{END}' if important_msg != '' else ''
    print(f"\n{dash_line}\n\n{message}{important_msg}\n\n{dash_line}\n")



from setuptools import setup, find_packages

with open("README.md", "r") as readme:
    long_description = readme.read()

with open("requirements.txt", "r") as req:
    install_requires = req.read().split("\n")


setup(
    name="fast-dataset-cleaner",
    version="1.0.0",
    author="Thomas Chabal",
    author_email="t.chabal@hotmail.fr",
    description="Web platform for fast image dataset cleaning",
    keywords="annotation,dataset cleaning,binary annotation,fast dataset cleaner,image dataset",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PhotoRoom/fast-dataset-cleaner",
    packages=find_packages(),
    include_package_data=True,
    scripts=['scripts/fast-dataset-cleaner'],
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import os

from ..image_builder import build_image


class ImageService:
    def __init__(self):
        self.img_dir = None
        self.mask_dir = None
        self.img_extension = ''
        self.mask_extension = ''
        self.use_masks = None

        
    def set_constants(self, img_folder, img_ext, mask_folder, mask_ext, with_masks):
        if img_ext is not None and img_ext != self.img_extension:
            self.img_extension = img_ext
    
        if mask_ext is not None and mask_ext != self.mask_extension:
            self.mask_extension = mask_ext
        
        if not os.path.exists(img_folder):
            return { 'error': f'Images folder not found: {img_folder}' }, 200
        if img_folder[-1] != '/':
            img_folder += '/'
        if img_folder != self.img_dir:
            self.img_dir = img_folder
            
        if with_masks:
            if mask_folder[-1] != '/':
                mask_folder += '/'
            if not os.path.exists(mask_folder):
                return { 'error': f'Masks folder not found: {mask_folder}' }, 200
            if mask_folder != self.mask_extension:
                self.mask_dir = mask_folder
                self.use_masks = with_masks


    def build_image(self, image_id):
        if image_id == '':
            return { 'error': 'Provide an image id' }, 200
        
        image_path = f'{self.img_dir}{image_id}{self.img_extension}'
        mask_path = f'{self.mask_dir}{image_id}{self.mask_extension}' if self.use_masks else None
        
        return build_image(image_path, mask_path)
